searchData={"items":[{"type":"module","title":"Instructor","doc":"Instructor.ex is a spiritual port of the great [Instructor Python Library](https://github.com/jxnl/instructor) by [@jxnlco](https://twitter.com/jxnlco).\nThis library brings structured prompting to LLMs. Instead of receiving text as output, Instructor will coax the LLM to output valid JSON that maps directly to the provided Ecto schema.\nIf the LLM fails to do so, or provides values that do not pass your validations, it will provide you utilities to automatically retry with the LLM to correct errors.\nBy default it's designed to be used with the [OpenAI API](https://platform.openai.com/docs/api-reference/chat-completions/create), however it provides an extendable adapter behavior to work with [ggerganov/llama.cpp](https://github.com/ggerganov/llama.cpp) and [Bumblebee (Coming Soon!)](https://github.com/elixir-nx/bumblebee).\n\nAt its simplest, usage is pretty straightforward,\n\n```elixir\ndefmodule SpamPredicition do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"","ref":"Instructor.html"},{"type":"module","title":"Field Descriptions: - Instructor","doc":"- class: Whether or not the email is spam\n  - reason: A short, less than 10 word rationalization for the classification\n  - score: A confidence score between 0.0 and 1.0 for the classification\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n\nis_spam? = fn text ->\n  Instructor.chat_completion(\n    model: \"gpt-3.5-turbo\",\n    response_model: SpamPredicition,\n    max_retries: 3,\n    messages: [\n      %{\n        role: \"user\",\n        content: \"\"\"\n        You purpose is to classify customer support emails as either spam or not.\n        This is for a clothing retailer business.\n        They sell all types of clothing.\n\n        Classify the following email: #{text}\n        \"\"\"\n      }\n    ]\n  )\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n\n# => {:ok, %SpamPredicition{class: :spam, reason: \"Nigerian prince email scam\", score: 0.98}}\n```\n\nSimply create an ecto schema, optionally provide a `@doc` to the schema definition which we pass down to the LLM, then make a call to `Instructor.chat_completion/1` with context about the task you'd like the LLM to complete.\nYou can also provide a `validate_changeset/1` function via the `use Instructor.Validator` which will provide a set of code level ecto changeset validations. You can use this in conjunction with `max_retries: 3` to automatically, iteratively go back and forth with the LLM up to `n` times with any validation errors so that it has a chance to fix them.\n\n**Curious to learn more? Unsure of how you'd use this? Check out our extensive set of [tutorials](#)**","ref":"Instructor.html#module-field-descriptions"},{"type":"module","title":"Configuration - Instructor","doc":"To configure the default OpenAI adapter you can set the configuration,\n\n```elixir\nconfig :openai, api_key: \"sk-........\"\nconfig :openai, http_options: [recv_timeout: 10 * 60 * 1000]\n```\n\nTo use a local LLM, you can install and run [llama.cpp serer](https://github.com/ggerganov/llama.cpp/blob/master/examples/server/README.md) and tell instructor to use it,\n\n```elixir\nconfig :instructor, adapter: Instructor.Adapters.Llamacpp\n```","ref":"Instructor.html#module-configuration"},{"type":"function","title":"Instructor.chat_completion/1","doc":"Create a new chat completion for the provided messages and parameters.\n\nThe parameters are passed directly to the LLM adapter.\nBy default they shadow the OpenAI API parameters.\nFor more information on the parameters, see the [OpenAI API docs](https://platform.openai.com/docs/api-reference/chat-completions/create).\n\nAdditionally, the following parameters are supported:\n\n  * `:response_model` - The Ecto schema to validate the response against.\n  * `:max_retries` - The maximum number of times to retry the LLM call if it fails, or does not pass validations.\n                     (defaults to `0`)","ref":"Instructor.html#chat_completion/1"},{"type":"function","title":"Examples - Instructor.chat_completion/1","doc":"iex> Instructor.chat_completion(%{\n  ...>   model: \"gpt-3.5-turbo\",\n  ...>   response_model: Instructor.Demos.SpamPrediction,\n  ...>   messages: [\n  ...>     %{\n  ...>       role: \"user\",\n  ...>       content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n  ...>     }\n  ...> })\n  {:ok,\n      %Instructor.Demos.SpamPrediction{\n          class: :spam\n          score: 0.999\n      }}","ref":"Instructor.html#chat_completion/1-examples"},{"type":"behaviour","title":"Instructor.Adapter","doc":"Behavior for `Instructor.Adapter`.","ref":"Instructor.Adapter.html"},{"type":"callback","title":"Instructor.Adapter.chat_completion/2","doc":"","ref":"Instructor.Adapter.html#c:chat_completion/2"},{"type":"module","title":"Instructor.Adapters.Llamacpp","doc":"Runs against the llama.cpp server. To be clear this calls the llamacpp specific\nendpoints, not the open-ai compliant ones.\n\nYou can read more about it here:\n  https://github.com/ggerganov/llama.cpp/tree/master/examples/server","ref":"Instructor.Adapters.Llamacpp.html"},{"type":"function","title":"Instructor.Adapters.Llamacpp.chat_completion/2","doc":"Run a completion against the llama.cpp server, not the open-ai compliant one.\nThis gives you more specific control over the grammar, and the ability to\nprovide other parameters to the specific LLM invocation.\n\nYou can read more about the parameters here:\n  https://github.com/ggerganov/llama.cpp/tree/master/examples/server","ref":"Instructor.Adapters.Llamacpp.html#chat_completion/2"},{"type":"function","title":"Examples - Instructor.Adapters.Llamacpp.chat_completion/2","doc":"iex> Instructor.chat_completion(%{\n  ...>   model: \"mistral-7b-instruct\",\n  ...>   messages: [\n  ...>     %{ role: \"user\", content: \"Classify the following text: Hello I am a Nigerian prince and I would like to send you money!\" },\n  ...>   ],\n  ...>   response_model: response_model,\n  ...>   temperature: 0.5,\n  ...> })","ref":"Instructor.Adapters.Llamacpp.html#chat_completion/2-examples"},{"type":"module","title":"Instructor.Adapters.OpenAI","doc":"Documentation for `Instructor.Adapters.OpenAI`.","ref":"Instructor.Adapters.OpenAI.html"},{"type":"module","title":"Instructor.JSONSchema","doc":"","ref":"Instructor.JSONSchema.html"},{"type":"function","title":"Instructor.JSONSchema.from_ecto_schema/1","doc":"Generates a JSON Schema from an Ecto schema.\n\n  Note: This will output a correct JSON Schema for the given Ecto schema, but\n  it will not necessarily be optimal, nor support all Ecto types.","ref":"Instructor.JSONSchema.html#from_ecto_schema/1"},{"type":"function","title":"Instructor.JSONSchema.title_for/1","doc":"","ref":"Instructor.JSONSchema.html#title_for/1"},{"type":"behaviour","title":"Instructor.Validator","doc":"By default you'll get whatever OpenAI returns.\nThis behavior provides a hook for you to critique the response using standard ecto changesets validations.\nThis can be used in conjuction with the `:max_retries` parameter to `Instructor.chat_completion/1` to retry the completion until it passes your validation.","ref":"Instructor.Validator.html"},{"type":"behaviour","title":"Examples - Instructor.Validator","doc":"defmodule Instructor.Demos.SpamPrediction do\n    use Ecto.Schema\n\n    @primary_key false\n    schema \"spam_prediction\" do\n        field :class, :string\n        field :score, :float\n    end\n\n    @impl true\n    def validate_changeset(changeset) do\n        changeset\n        |> validate_number(:score, less_than_or_equal_to: 1.0)\n    end\n  end\n\n  iex> Instructor.chat_completion(%{\n  ...>   model: \"gpt-3.5-turbo\",\n  ...>   response_model: Instructor.Demos.SpamPrediction,\n  ...>   max_retries: 1,\n  ...>   messages: [\n  ...>     %{\n  ...>       role: \"user\",\n  ...>       content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n  ...>     }\n  ...> })\n  {:error, %Ecto.Changeset{\n      action: nil,\n      changes: %{},\n      errors: [\n          score: {\"is invalid\", [validation: :number, validation_opts: [less_than_or_equal_to: 1.0]]}\n      ],\n      data: %Instructor.Demos.SpamPrediction{\n          class: nil,\n          score: nil\n      },\n      valid?: false\n  }}","ref":"Instructor.Validator.html#module-examples"},{"type":"callback","title":"Instructor.Validator.validate_changeset/1","doc":"","ref":"Instructor.Validator.html#c:validate_changeset/1"},{"type":"extras","title":"instructor_ex","doc":"# instructor_ex\n\n_Structured, Ecto outputs with OpenAI (and OSS LLMs)_\n\n---\n\n[![KinoShell version](https://img.shields.io/hexpm/v/instructor.svg)](https://hex.pm/packages/instructor)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg)](https://hexdocs.pm/instructor/)\n[![GitHub stars](https://img.shields.io/github/stars/thmsmlr/instructor_ex.svg)](https://github.com/thmsmlr/instructor_ex/stargazers)\n[![Twitter Follow](https://img.shields.io/twitter/follow/thmsmlr?style=social)](https://twitter.com/thmsmlr)\n\n<!-- Docs -->\n\nInstructor.ex is a spiritual port of the great [Instructor Python Library](https://github.com/jxnl/instructor) by [@jxnlco](https://twitter.com/jxnlco).\nThis library brings structured prompting to LLMs. Instead of receiving text as output, Instructor will coax the LLM to output valid JSON that maps directly to the provided Ecto schema.\nIf the LLM fails to do so, or provides values that do not pass your validations, it will provide you utilities to automatically retry with the LLM to correct errors.\nBy default it's designed to be used with the [OpenAI API](https://platform.openai.com/docs/api-reference/chat-completions/create), however it provides an extendable adapter behavior to work with [ggerganov/llama.cpp](https://github.com/ggerganov/llama.cpp) and [Bumblebee (Coming Soon!)](https://github.com/elixir-nx/bumblebee).\n\nAt its simplest, usage is pretty straightforward,\n\n```elixir\ndefmodule SpamPredicition do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"","ref":"readme.html"},{"type":"extras","title":"Field Descriptions: - instructor_ex","doc":"- class: Whether or not the email is spam\n  - reason: A short, less than 10 word rationalization for the classification\n  - score: A confidence score between 0.0 and 1.0 for the classification\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n\nis_spam? = fn text ->\n  Instructor.chat_completion(\n    model: \"gpt-3.5-turbo\",\n    response_model: SpamPredicition,\n    max_retries: 3,\n    messages: [\n      %{\n        role: \"user\",\n        content: \"\"\"\n        You purpose is to classify customer support emails as either spam or not.\n        This is for a clothing retailer business.\n        They sell all types of clothing.\n\n        Classify the following email: #{text}\n        \"\"\"\n      }\n    ]\n  )\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n\n# => {:ok, %SpamPredicition{class: :spam, reason: \"Nigerian prince email scam\", score: 0.98}}\n```\n\nSimply create an ecto schema, optionally provide a `@doc` to the schema definition which we pass down to the LLM, then make a call to `Instructor.chat_completion/1` with context about the task you'd like the LLM to complete.\nYou can also provide a `validate_changeset/1` function via the `use Instructor.Validator` which will provide a set of code level ecto changeset validations. You can use this in conjunction with `max_retries: 3` to automatically, iteratively go back and forth with the LLM up to `n` times with any validation errors so that it has a chance to fix them.\n\n**Curious to learn more? Unsure of how you'd use this? Check out our extensive set of [tutorials](#)**","ref":"readme.html#field-descriptions"},{"type":"extras","title":"Configuration - instructor_ex","doc":"To configure the default OpenAI adapter you can set the configuration,\n\n```elixir\nconfig :openai, api_key: \"sk-........\"\nconfig :openai, http_options: [recv_timeout: 10 * 60 * 1000]\n```\n\nTo use a local LLM, you can install and run [llama.cpp serer](https://github.com/ggerganov/llama.cpp/blob/master/examples/server/README.md) and tell instructor to use it,\n\n```elixir\nconfig :instructor, adapter: Instructor.Adapters.Llamacpp\n```\n\n<!-- Docs -->","ref":"readme.html#configuration"},{"type":"extras","title":"Installation - instructor_ex","doc":"In your mix.exs,\n\n```elixir\ndef deps do\n  [\n    {:instructor, \"~> 0.0.2\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"TODO - instructor_ex","doc":"- [x] Tests\n    - [x] JSONSchema\n    - [x] gbnf\n- [x] Add JSONSchema --> GBNF computation\n- [x] Add field descriptions\n- [x] Add validators\n- [x] Add typespecs and docs\n- [x] README, hexdocs, and hex.pm should have same README example\n- [ ] GBNF should enforce required properties on objects, currently they're optional.\n- [ ] GBNF limit the number of digits in number tokens -- small models can sometimes run off to infinit digits\n- [ ] Better support nullable\n- [x] Retry Logic\n- [ ] llamacpp adapter broken, needs to support openai input/output API\n- [ ] Add `llm_validator`\n- [ ] Support binaries and binary_id in JSONSchema and GBNF\n- [x] Verify :naive_datetime support\n- [ ] Support :binary_id\n- [x] Support OpenAI Tools AP\n- [ ] Logging for Distillation / Finetuning\n- [ ] Add a Bumblebee adapter\n- [ ] Add Livebook Tutorials, include in Hexdocs\n    - [ ] Text Classification\n    - [ ] Self Critique\n    - [ ] Image Extracting Tables\n    - [ ] Moderation\n    - [ ] Citations\n    - [ ] Knowledge Graph\n    - [ ] Entity Resolution\n    - [ ] Search Queries\n    - [ ] Query Decomposition\n    - [ ] Recursive Schemas\n    - [ ] Table Extraction\n    - [ ] Action Item and Dependency Mapping\n    - [ ] Multi-File Code Generation\n    - [ ] PII Data Sanitization\n- [x] Update hexdocs homepage to include example for tutorial\n- [ ] Setup Github CI for testing, add badge to README","ref":"readme.html#todo"},{"type":"extras","title":"Blog Posts - instructor_ex","doc":"- [ ] Why structured prompting?\n\n    Meditations on new HCI.\n    Finally we have software that can understand text. f(text) -> text.\n    This is great, as it gives us a new domain, but the range is still text.\n    While we can use string interpolation to map Software 1.0 into f(text), the outputs are not interoperable with Software 1.0.\n    Hence why UXs available to us are things like Chatbots as our users have to interpret the output.\n\n    Instructor, structure prompting, gives use f(text) -> ecto_schema.\n    Schemas are the lingua franca of Software 1.0.\n    With Instrutor we can now seamlessly move back and forth between Software 1.0 and Software 2.0.\n\n    Now we can maximally leverage AI...\n\n- [ ] From GPT-4 to zero-cost production - Distilation, local-llms, and the cost structure of AI.\n\n    ... 😘","ref":"readme.html#blog-posts"},{"type":"extras","title":"Tutorial","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Tutorial\n\n```elixir\nMix.install(\n  [\n    # {:instructor, \"~> 0.0.2\"}\n    {:instructor, path: \"/Users/thomas/code/instructor_ex\"}\n  ],\n  config: [\n    instructor: [\n      adapter: Instructor.Adapters.OpenAI\n    ],\n    openai: [\n      api_key: System.fetch_env!(\"LB_OPENAI_API_KEY\"),\n      http_options: [recv_timeout: 10 * 60 * 1000]\n    ]\n  ]\n)\n```","ref":"tutorial.html"},{"type":"extras","title":"Introduction - Tutorial","doc":"Instructor is a library to do structured prompting with OpenAI and open source LLMs. While the idea is pretty simple, through this and the other examples you'll realize how powerful a concept this is.\n\n**So first off, what is structure prompting?**\n\nWhat if the LLM returned data conforming to a complicated nested schema that your code knows how to work with? Well, that's structure prompting. It's a way of cohercing the LLM to producing it's response in a known format that your downstream code can handle. In the case of Instructor, we use Ecto to provide those schemas. Good old Ecto, something you're already familiar with.\n\nSo, without further ado, let's take define a schema and take it for a spin!\n\n```elixir\ndefmodule Politician do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"\n  A description of United States Politicians and the offices that they held,","ref":"tutorial.html#introduction"},{"type":"extras","title":"Fields: - Tutorial","doc":"- first_name: Their first name\n  - last_name: Their last name\n  - offices_held:\n    - office: The branch and position in government they served in\n    - from_date: When they entered office or null\n    - until_date: The date they left office or null\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:first_name, :string)\n    field(:last_name, :string)\n\n    embeds_many :offices_held, Office, primary_key: false do\n      field(:office, Ecto.Enum,\n        values: [:president, :vice_president, :govenor, :congress, :senate]\n      )\n\n      field(:from_date, :date)\n      field(:to_date, :date)\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Politician, <<70, 79, 82, 49, 0, 0, 17, ...>>,\n [__schema__: 1, __schema__: 1, __schema__: 1, __schema__: 1, __schema__: 2, __schema__: 2, ...]}\n```\n\nGreat, we have our schema describing politicans and the offices they held. Let's notice a few things that may stand out from regular Ecto usage. First, since there is no database backing the schema, it doesn't make sense to give it a primary_key. This also makes sense because there is no sensible value for the LLM to respond with.\n\nAlso we use a `@doc` on the schema. This isn't just for documentation purposes of the tutorial. Instructor will take any `@doc` tag and provide it to the LLM. Generally you'll want to use this to provide semantic descriptions of the fields and general context to the LLM to ensure you get the outputs you want. In our case we want to push the LLM to understand that we are only considering American politicians.\n\nSo, let's try asking the LLM to give us some politicians.\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-3.5-turbo\",\n  response_model: Politician,\n  messages: [\n    %{\n      role: \"user\",\n      content:\n        \"Who won the American 2020 election and what offices have they held over their career?\"\n    }\n  ]\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:error,\n #Ecto.Changeset ,\n         valid?: false\n       >\n     ]\n   },\n   errors: [],\n   data: #Politician<>,\n   valid?: false\n >}\n```\n\nAmazing, right? Using nothing more than one of the top libraries in Elixir, Ecto, we were able to get structured output from our LLM. The data returned is ready to be processed by our regular Elixir code. Instructor supports all field types that you can express in Ecto, including embedded and associated schemas.\n\nIt's almost as if the LLM inputted the data into a Phoenix Form. All the utilities that you use to process that kind of data, you can use to process the outputs of Instructor.\n\nOne of the superpowers of this is that since we're just using changesets under the hood, you can use the same validations that you would use elsewhere in your app. Let's look at that in the next section.","ref":"tutorial.html#fields"},{"type":"extras","title":"Validations - Tutorial","doc":"Instructor provides a lightweight behavior where you can define a callback function that we will call to validate the data returned by the LLM using Ecto changesets. There is nothing fancy to this API. It's just a changeset in and a changeset out.\n\n```elixir\ndefmodule NumberSeries do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @primary_key false\n  embedded_schema do\n    field(:series, {:array, :integer})\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_length(:series, min: 10)\n    |> Ecto.Changeset.validate_change(:series, fn\n      field, values ->\n        if Enum.sum(values) |> rem(2) == 0 do\n          []\n        else\n          [{field, \"The sum of the series must be even\"}]\n        end\n    end)\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, NumberSeries, <<70, 79, 82, 49, 0, 0, 18, ...>>, {:validate_changeset, 1}}\n```\n\nIn this albeit contrived example, we're going to get the LLM to return a series of numbers and validate whether it has at least 10 numbers and that the sum of the series is even.\n\nWhen we ask for fewer than ten numbers, Instructor will return an error tuple with a change set that is invalid.\n\n```elixir\n{:error, changeset} =\n  Instructor.chat_completion(\n    model: \"gpt-3.5-turbo\",\n    response_model: NumberSeries,\n    messages: [\n      %{role: \"user\", content: \"Give me the first 5 integers\"}\n    ]\n  )\n\n# Render our the errors down to strings.\nerrors =\n  Ecto.Changeset.traverse_errors(changeset, fn {msg, opts} ->\n    Regex.replace(~r\"%{(\\w+)}\", msg, fn _, key ->\n      opts |> Keyword.get(String.to_existing_atom(key), key) |> to_string()\n    end)\n  end)\n\n{changeset.changes, errors}\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{%{series: [1, 2, 3, 4, 5]},\n %{series: [\"The sum of the series must be even\", \"should have at least 10 item(s)\"]}}\n```\n\nNow the beauty of this is that since we have human readable errors from our validations, we can just turn around and pass those back into the LLM to get it to fix its own errors.\n\nInstructor provides a convenience parameter, `max_retries` for you in the initial call which will retry against the validations up to n times.\n\n```elixir\nInstructor.chat_completion(\n  model: \"gpt-3.5-turbo\",\n  response_model: NumberSeries,\n  max_retries: 10,\n  messages: [\n    %{role: \"user\", content: \"Give some random integers\"}\n  ]\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\n21:36:54.810 [debug] Retrying LLM call for Elixir.NumberSeries...\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok, %NumberSeries{series: [2, 4, 6, 8, 10, 12, 14, 16, 18, 20]}}\n```\n\n<!-- livebook:{\"offset\":7027,\"stamp\":{\"token\":\"XCP.K4lM2RF9aHbWXus5rF49n4QOEk359bZ6KzBPROMyev_3KLxFvOL3RoYkEhdybgzOaRBqv0gQ1IkXBfstKZ7TqXe69unuaqhAU_9anxT6wXJSiqa8-UtQFA8\",\"version\":2}} -->","ref":"tutorial.html#validations"}],"content_type":"text/markdown"}